// Turbo51-Unit
// Driver for I2C-Bus
// J. Rathlev, Jan. 2011

unit I2C;

{$C- }
interface

type
  TBytePointer = ^byte;
  TI2CBuffer = array [0..7] of byte;

var  
  SSCL : boolean absolute P3.5; volatile;
  SSDA : boolean absolute P3.4; volatile;
  
const
// Port-Expander PCF8574
  PortExpander0 = $40;        // I2C Base address of PCF8574
  PortExpander1 = $70;        // I2C Base address of PCF8574A
         
// AD/DA converter PCF8591
  ADConverter  = $90;        // I2C Base address of PCF8591

// Temperature sensor LM 75
  TempSensor  = $90;        //I2C Base address of LM 75

// write bytes from buffer to EEPROM
function I2CWriteBlock (Addr,NBytes : byte; Buffer : TBytePointer) : boolean;

// read bytes from EEPROM to buffer            
function I2CReadBlock (Addr,NBytes : byte; Buffer : TBytePointer) : boolean;

// write bytes to device
function I2CWriteDevice (Device,Addr,NBytes : byte; Buffer : TBytePointer) : boolean;  

// read bytes from device           
function I2CReadDevice (Device,Addr,NBytes : byte; Buffer : TBytePointer) : boolean;

implementation

const
// EEPROMs  (24C04,..)
  EPRADR =  $A;        // I2C-Address of EEPROM
  TryCnt = 5;

// generate start condition  
function I2CStart : boolean; Assembler;
Asm
    SETB    SSCL            // SCL = H
    NOP
    SETB    SSDA            // SDA = H
    NOP
    MOV     R2,#TryCnt      // 20s prfen und warten
    CLR     C               // Fehler
@I2CSTW: 
    JNB     SSCL,@I2CSTE    // Fehler: SCL = L (2 s)
    JNB     SSDA,@I2CSTE    // Fehler: SDA = L (2 s)
    DJNZ    R2,@I2CSTW      // warten          (2 s)    
    CLR     SSDA            // SDA = L
    NOP                                   
    CLR     SSCL            // SCL = L
    SETB    C               // ok
@I2CSTE:    
  end;  

// generate stop condition  
procedure I2CStop; Assembler;
Asm
    CLR     SSDA            // SDA = L
    NOP
    SETB    SSCL            // SCL = H
    NOP
    SETB    SSDA            // SDA = H
  end;  
  
// write bit from C to bus  
function I2CWriteBit : boolean; Assembler;
Asm
    MOV     SSDA,C
    NOP
    JNC     @I2CW1          // Bit = L ?
    MOV     R2,#TryCnt      // ja - prfen, ob Bus frei
@I2CWW1: 
    JB      SSDA,@I2CW1     // SDA = H ?
    DJNZ    R2,@I2CWW1      // nein, warten
    SETB    SSCL            // Bus belegt, freigeben
    CLR     C
    SJMP    @I2CWE          // Fehler
@I2CW1:  
    SETB    SSCL            // SCL = H
    NOP
    MOV     R2,#TryCnt      // ja - prfen, ob Bus frei
@I2CWW2: 
    JB      SSCL,@I2CW2     // SCL = H ?
    DJNZ    R2,@I2CWW2      // nein, warten
    SETB    SSDA            // Bus belegt, freigeben
    CLR     C
    SJMP    @I2CWE          // Fehler
@I2CW2:  
    CLR     SSCL            // SCL = L
    SETB    C               // ok
@I2CWE:  
  end;

// read bit from bus to F0
function I2CReadBit : boolean; Assembler;
Asm
    SETB    SSDA            // SDA = H
    NOP
    SETB    SSCL            // SCL = H
    NOP
    MOV     R2,#TryCnt      // prfen, ob Bus frei
@I2CRW1: 
    JB      SSCL,@I2CR1     // SCL = H ?
    DJNZ    R2,@I2CRW1      // nein, warten
    CLR     C               
    SJMP    @I2CRE           // Fehler
@I2CR1:  
    MOV     C,SSDA          // lese Bit
    MOV     F0,C
    CLR     SSCL            // SCL = L
    SETB    C               // ok
@I2CRE:  
  end;
  
// write byte to bus
function I2CWriteByte (AByte : byte) : boolean; Assembler;
Asm
    MOV     A,AByte
    MOV     R3,#8            // Zhler
@I2CWB1: 
    RLC     A                // Bits nach C
    LCALL   I2CWriteBit      // Bit schreiben
    JNC     @I2CWBE          // Fehler
    DJNZ    R3,@I2CWB1       // Schleife fr 8 Bits
    LCALL   I2CReadBit       // Besttigungsbit (F0) lesen
    JNC     @I2CWBE          // Fehler
    MOV     C,F0             // Besttigungsbit 
    CPL     C
    JC      @I2CWBE          // ok
    LCALL   I2CSTop          // Stop
@I2CWBE: 
  end;
          
// read byte from bus
function I2CReadByte (var AByte : byte; Ack : boolean) : boolean; Assembler;
Asm
    MOV     R3,#8            // Zhler
@I2CRB1: 
    LCALL   I2CReadBit       // Bit lesen
    JNC     @I2CRBE          // Fehler
    MOV     C,F0
    RLC     A                // Bit nach A
    DJNZ    R3,@I2CRB1       // Schleife fr 8 Bits
    MOV     R0,AByte
    MOV     @R0,A            // store byte
    MOV     C,Ack          
    CPL     C
    LCALL   I2CWriteBit      // set ACKK
    SETB    SSDA             // SDA = H    
@I2CRBE:   
  end;
          
// write bytes from buffer to EEPROM
function I2CWriteBlock (Addr,NBytes : byte; Buffer : TBytePointer) : boolean;
begin
  Result:=false;
  if (NBytes<>0) and I2CStart then begin
    if I2CWriteByte(EPRADR and $FE) then begin // Base address and bank
      if I2CWriteByte(Addr) then repeat
        Result:=I2CWriteByte(Buffer^);
        inc(Buffer);
        dec(NBytes);
        until not Result or (Nbytes=0);
      end;
    end;
  I2CStop;
  end;
   
// read bytes from EEPROM to buffer            
function I2CReadBlock (Addr,NBytes : byte; Buffer : TBytePointer) : boolean;
begin
  Result:=false;
  if (NBytes<>0) and I2CStart then begin
    if I2CWriteByte(EPRADR and $FE) then begin // Base address and bank
      if I2CWriteByte(Addr) then begin    
        if I2CStart then begin
          if I2CWriteByte((EPRADR and $FE) or 1) then begin
            while NBytes>1 do begin
              Result:=I2CReadByte(Buffer^,true);
              inc(Buffer);
              dec(NBytes);
              if not Result then Break;
              end;
            Result:=Result and I2CReadByte(Buffer^,false);  // no Ack. on last byte
            end;
          end;
        end;
      end;
    end;
  I2CStop;
  end;

// write bytes to device
function I2CWriteDevice (Device,Addr,NBytes : byte; Buffer : TBytePointer) : boolean;  
begin
  Result:=false;
  if I2CStart then begin
    if I2CWriteByte(Device or (Addr and 7) shl 1) then repeat
      Result:=I2CWriteByte(Buffer^);
      inc(Buffer);
      dec(NBytes);
      until not Result or (Nbytes=0);
    end; 
  I2CStop;
  end;
   
// read bytes from device           
function I2CReadDevice (Device,Addr,NBytes : byte; Buffer : TBytePointer) : boolean;
begin
  Result:=false;
  if I2CStart then begin
    if I2CWriteByte(Device or ((Addr and 7) shl 1) or 1) then begin
      Result:=true;
      while NBytes>1 do begin
        Result:=I2CReadByte(Buffer^,true);
        inc(Buffer);
        dec(NBytes);
        if not Result then Break;
        end;
      Result:=Result and I2CReadByte(Buffer^,false);  // no Ack. on last byte
      end;
    end; 
  I2CStop;
  end;
  
end.